from abc import *
from ntt_signal import Signal


class IModel(ABC):
    @abstractmethod
    def _AppendModelData(self, strAttributeName: str, bIsModel: bool = False) -> None:
        pass

    @abstractmethod
    def ToDict(self) -> dict:
        pass

    @abstractmethod
    def FromDict(self, dictData: dict) -> None:
        pass

    @abstractmethod
    def Attach(self, signal: Signal) -> None:
        pass

    @abstractmethod
    def ToFile(self, strFileName: str) -> None:
        pass

    @abstractmethod
    def FromFile(self, strFileName: str) -> None:
        pass


from typing import Any
from ntt_signal import Signal
from typing import *
from ..IModel import IModel 
from .Property import Property
from ntt_observable_list import ObservableList


class ListProperty(Property):
    def __init__(self, model: IModel, lstData: List[Any], strAttributeName: str) -> None:
        super().__init__(model, strAttributeName, list)

        self._lstData = ObservableList(lstData)
        self._lstData.Attach(self)

    def GetValue(self) -> IModel:
        return self._lstData

    def SetValue(self, lstData: List[Any], bNotify: bool = True) -> None:
        self._lstData = ObservableList(lstData)
        self._lstData.Attach(self)


import unittest
from unittest.mock import Mock
from ntt_json_model import *
from .TestClass import *


class ModelPropertyTest(unittest.TestCase):
    model = None
    def setUp(self) -> None:
        self.model = TestModelClass() 

    def test_WhenChangeThePropertyOfTheModelAttribute_ThenTheModelIsEmitted(self):
        testCallback = Mock()

        self.model.Connect(testCallback)

        self.model.TestModel.Name = "ThaoNguyenThe"

        testCallback.assert_called_once()

    def test_WhenExportDataOfModelPropertyToDict_ThenGetRightValue(self):
        self.model.Score = 4
        self.model.TestModel.Name = "ThaoNguyenThe"
        self.model.TestModel.Temp = 3.2
        self.model.TestModel.Scores.extend([3, 2])

        dictData: dict = self.model.ToDict()

        self.assertDictEqual(
            dictData,
            {
                "__class__": "TestModelClass",
                "_nScore": 4,
                "_mTestModel": {
                    "__class__": "TestModel",
                    "_nValue": 3,
                    "_fTemp": 3.2,
                    "_strName": "ThaoNguyenThe",
                    "_lstScores": [3, 2]
                }
            }
        )

    def test_WhenLoadDataModelPropertyFromDict_ThenGetRightValueAndTheModelIsEmittedOnce(self):
        testCallback = Mock()
        self.model.Connect(testCallback)

        dictData = {
            "__class__": "TestModelClass",
            "_nScore": 4,
            "_mTestModel": {
                "__class__": "TestModel",
                "_nValue": 3,
                "_fTemp": 3.2,
                "_strName": "ThaoNguyenThe",
                "_lstScores": [3, 2]
            }
        }

        self.model.FromDict(dictData)

        self.assertEqual(self.model.Score, 4)
        self.assertEqual(self.model.TestModel.Name, "ThaoNguyenThe")
        self.assertEqual(self.model.TestModel.Value, 3)
        self.assertEqual(self.model.TestModel.Temp, 3.2)
        self.assertEqual(self.model.TestModel.Scores, [3, 2])
        testCallback.assert_called_once()


from typing import *
from abc import *
from ntt_signal import Signal
from ..IModel import IModel


class Property(Signal, ABC):
    def __init__(self, model:IModel, strAttributeName: str, 
                    datatype: type = None, 
                    bIsModel: bool = False) -> None:
        super().__init__(datatype)
        self.Attach(model)
        setattr(model, strAttributeName, self)
        model._AppendModelData(strAttributeName, bIsModel=bIsModel)

    @abstractmethod
    def GetValue(self) -> Any:
        pass

    @abstractmethod
    def SetValue(self, newValue: object, bNotify: bool = True) -> None:
        pass


import unittest
from typing import *
from unittest.mock import Mock
from .TestClass import *
from ntt_json_model import *


class ModelTest(unittest.TestCase):
    model = None

    def setUp(self) -> None:
        self.model = TestModel()

    def test_GivenModelWithData_WhenThatDataHasChanged_ThenTheModelIsEmitted(self):
        testCallback = Mock()
        self.model.Connect(testCallback)

        self.model.Value = 5

        testCallback.assert_called_once()

    def test_GivenModelWithData_WhenThatStrDataHasChanged_ThenTheModelIsEmitted(self):
        testCallback = Mock()
        self.model.Connect(testCallback)

        self.model.Temp = 5.3

        testCallback.assert_called_once()

    def test_WhenChangeFloatPropertyWithUnchangedData_ThenNoEmitted(self):
        testCallback = Mock()
        self.model.Connect(testCallback)

        self.model.Temp = 3.1
        
        testCallback.assert_not_called()

    def test_WhenChangeFloatPropertyWithInt_ThenModelIsEmitted(self):
        testCallback = Mock()
        self.model.Connect(testCallback)

        self.model.Temp = 5

        testCallback.assert_called_once()

    def test_WhenChangeDataWithoutChangingValueActually_ThenNoEmitted(self):
        testCallback = Mock()
        self.model.Connect(testCallback)

        self.model.Value = 3

        testCallback.assert_not_called()

    def test_WhenChangeStrPropertyWithInt_ThenModelIsEmitted(self):
        testCallback = Mock()
        self.model.Connect(testCallback)

        self.model.Name = "Test"

        testCallback.assert_called_once()

    def test_WhenChangeStrDataWithoutChangingValueActually_ThenNoEmitted(self):
        testCallback = Mock()
        self.model.Connect(testCallback)

        self.model.Name = "Hello"

        testCallback.assert_not_called()

    def test_WhenChangeListPropertyWithInt_ThenModelIsEmitted(self):
        testCallback = Mock()
        self.model.Connect(testCallback)

        self.model.Scores.append(4)

        testCallback.assert_called_once()

    def test_GivenModel_WhenExtractToDict_ThenDataConvertedToDict(self):
        self.model.Name = "Testing"
        self.model.Value = 3
        self.model.Temp = 45
        self.model.Scores.append(4)

        dictData = self.model.ToDict()

        self.assertDictEqual(
            dictData,
            {
                "__class__": "TestModel",
                "_nValue": 3,
                "_fTemp": 45.0,
                "_strName": "Testing",
                "_lstScores": [4]
            }
        )

    def test_GivenModel_WhenLoadDataFromDict_ModelIsEmittedOnce(self):
        dictData = {
            "__class__": "TestModel",
            "_nValue": 3,
            "_fTemp": 45.0,
            "_strName": "Testing",
            "_lstScores": [4],
        }
        testCallback = Mock()
        self.model.Connect(testCallback)

        self.model.FromDict(dictData)

        self.assertEqual(self.model.Name, "Testing")
        self.assertEqual(self.model.Temp, 45.0)
        self.assertEqual(self.model.Value, 3)
        self.assertListEqual(self.model.Scores, [4])
        testCallback.assert_called_once()

    def test_GivenModelWhichIsLoadedFromDict_WhenChangeTheProperty_ThenModelIsEmitted(self):
        dictData = {
            "__class__": "TestModel",
            "_nValue": 3,
            "_fTemp": 45.0,
            "_strName": "Testing",
            "_lstScores": [4],
        }
        testCallback = Mock()
        self.model.Connect(testCallback)

        self.model.FromDict(dictData)

        self.model.Value = 2
        self.model.Scores.clear()

        self.assertEqual(
            testCallback.call_count,
            3,
            "The Callback should be called 3 times",
        )

    def test_WhenLoadModelFromNonEnoughDataDict_ThenNoErrors(self):
        dictData = {
            "__class__": "TestModel",
            "_nValue": 3,
            "_strName": "Testing",
            "_lstScores": [4],
        }

        self.model.FromDict(dictData)

        self.assertEqual(self.model.Temp, 3.1)

    def test_WhenLoadModelFromNonClassKey_ThenRaiseError(self):
        dictData = {
            "_nValue": 3,
            "_strName": "Testing",
            "_fTemp": 4.1,
            "_lstScores": [4],
        }

        with self.assertRaises(InputDictError) as context:
            self.model.FromDict(dictData)
            
        self.assertEqual(str(context.exception), "Input dict has no __class__ key")

    def test_WhenLoadModelFromWrongClassValue_ThenRaiseError(self):
        dictData = {
            "__class__": "TestingModel",
            "_nValue": 3,
            "_fTemp": 45.0,
            "_strName": "Testing",
            "_lstScores": [4],
        }

        with self.assertRaises(TargetClassError) as context:
            self.model.FromDict(dictData)

        self.assertEqual(str(context.exception), "Expect TestModel but received TestingModel")


import unittest
import json
from unittest.mock import *
from ntt_json_model import *
from .TestClass import *


class ModelListPropertyTest(unittest.TestCase):
    model = None

    def setUp(self) -> None:
        self.model = TestModelListClass()

    def test_WhenChangeThePropertyOfTheModelAttribute_ThenTheModelIsEmitted(self):
        testCallback = Mock()

        self.model.Connect(testCallback)

        self.model.Models.extend([3, 4, 5])

        testCallback.assert_called_once()

    def test_WhenExportDataOfModelListPropertyToDict_ThenGetRightValue(self):
        self.model.Score = 4.3
        self.model.Models.extend([
            TestModelClass(nScore=4, nValue=3, fTemp=3.2, strName="ThaoNguyenThe"),
            TestModelClass(nScore=5, nValue=3, fTemp=3.2, strName="ThaoNguyenThe", lstScores=[3, 2]),
        ])

        dictData: dict = self.model.ToDict()

        self.assertDictEqual(
            dictData,
            {
                "__class__": "TestModelListClass",
                "_fScore": 4.3,
                "_mModels": [
                    {
                        "__class__": "TestModelClass",
                        "_nScore": 4,
                        "_mTestModel": {
                            "__class__": "TestModel",
                            "_nValue": 3,
                            "_fTemp": 3.2,
                            "_strName": "ThaoNguyenThe",
                            "_lstScores": []
                        }
                    },
                    {
                        "__class__": "TestModelClass",
                        "_nScore": 5,
                        "_mTestModel": {
                            "__class__": "TestModel",
                            "_nValue": 3,
                            "_fTemp": 3.2,
                            "_strName": "ThaoNguyenThe",
                            "_lstScores": [3, 2]
                        }
                    }
                ]
            }
        )

    def test_WhenLoadDataForModelListFromDict_ThenGetRightvalue(self):
        dictData = {
            "__class__": "TestModelListClass",
            "_fScore": 4.3,
            "_mModels": [
                {
                    "__class__": "TestModelClass",
                    "_nScore": 4,
                    "_mTestModel": {
                        "__class__": "TestModel",
                        "_nValue": 3,
                        "_fTemp": 3.2,
                        "_strName": "ThaoNguyenThe",
                        "_lstScores": []
                    }
                },
                {
                    "__class__": "TestModelClass",
                    "_nScore": 5,
                    "_mTestModel": {
                        "__class__": "TestModel",
                        "_nValue": 3,
                        "_fTemp": 3.2,
                        "_strName": "ThaoNguyenThe",
                        "_lstScores": [3, 2]
                    }
                }
            ]
        }
        testCallback = Mock()
        self.model.Connect(testCallback)

        self.model.FromDict(dictData)

        self.assertEqual(self.model.Score, 4.3)
        self.assertEqual(len(self.model.Models), 2)
        self.assertEqual(self.model.Models[0].Score, 4)
        self.assertEqual(self.model.Models[1].Score, 5)
        self.assertEqual(self.model.Models[0].TestModel.Name, "ThaoNguyenThe")
        self.assertEqual(self.model.Models[1].TestModel.Name, "ThaoNguyenThe")
        self.assertListEqual(self.model.Models[0].TestModel.Scores, [])
        self.assertListEqual(self.model.Models[1].TestModel.Scores, [3, 2])

        testCallback.assert_called_once()

    def test_WhenLoadDataForModelListFromDictAndChangeList_TheModelIsEmitted(self):
        dictData = {
            "__class__": "TestModelListClass",
            "_fScore": 4.3,
            "_mModels": [
                {
                    "__class__": "TestModelClass",
                    "_nScore": 4,
                    "_mTestModel": {
                        "__class__": "TestModel",
                        "_nValue": 3,
                        "_fTemp": 3.2,
                        "_strName": "ThaoNguyenThe",
                        "_lstScores": []
                    }
                },
                {
                    "__class__": "TestModelClass",
                    "_nScore": 5,
                    "_mTestModel": {
                        "__class__": "TestModel",
                        "_nValue": 3,
                        "_fTemp": 3.2,
                        "_strName": "ThaoNguyenThe",
                        "_lstScores": [3, 2]
                    }
                }
            ]
        }
        testCallback = Mock()
        self.model.FromDict(dictData)
        self.model.Connect(testCallback)

        self.model.Models.append(TestModel())

        testCallback.assert_called_once()


from typing import Any
from ntt_signal import Signal
from .Property import Property


class StrProperty(Property):
    def __init__(self, model: Signal, strValue: str, strAttributeName: str) -> None:
        super().__init__(model, strAttributeName, str)

        self._strValue = strValue

    def GetValue(self) -> Any:
        return self._strValue

    def SetValue(self, strValue: str, bNotify: bool = True) -> None:
        if self._strValue != strValue:
            self._strValue = strValue
            if bNotify:
                self.Emit(self._strValue)


class InputDictError(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

class TargetClassError(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


from typing import Any
from ntt_signal import Signal
from .Property import Property


class IntegerProperty(Property):
    def __init__(self, model: Signal, nValue: int, strAttributeName: str) -> None:
        super().__init__(model, strAttributeName, int)
        self._nValue = nValue

    def GetValue(self) -> Any:
        return self._nValue
    
    def SetValue(self, nNewValue:int, bNotify: bool = True) -> None:
        if nNewValue != self._nValue:
            self._nValue = nNewValue
            if bNotify:
                self.Emit(self._nValue)


from typing import *
from typing import Any
from ..IModel import IModel
from .Property import Property
from ntt_observable_list import ObservableList


class ModelListProperty(Property):
    def __init__(self, model: IModel, mModels: List[IModel], strAttributeName: str) -> None:
        super().__init__(model, strAttributeName, None, bIsModel=True)

        self._mModels = ObservableList(mModels)
        self._mModels.Attach(self)

    def GetValue(self) -> ObservableList:
        return self._mModels
    
    def SetValue(self, mNewModels: ObservableList[IModel], bNotify: bool = True) -> None:
        self._mModels = mNewModels
        self._mModels.Attach(self)


from ntt_json_model import *


class TestModel(ModelBase):
    def __init__(self, 
                    nValue: int = 3, 
                    strName: str = "Hello",
                    fTemp: float = 3.1,
                    lstScores: List[int] = []) -> None:
        super().__init__()

        IntegerProperty(self, nValue, "_nValue")
        FloatProperty(self, fTemp, "_fTemp")
        StrProperty(self, strName, "_strName")
        ListProperty(self, lstScores, "_lstScores")

    @property
    def Value(self) -> int:
        return self._nValue.GetValue()

    @Value.setter
    def Value(self, nValue: int) -> None:
        self._nValue.SetValue(nValue)

    @property
    def Temp(self) -> str:
        return self._fTemp.GetValue()

    @Temp.setter
    def Temp(self, fTemp: float) -> None:
        self._fTemp.SetValue(fTemp)
    
    @property
    def Name(self) -> str:
        return self._strName.GetValue()

    @Name.setter
    def Name(self, strName: str) -> None:
        self._strName.SetValue(strName)

    @property
    def Scores(self) -> List[int]:
        return self._lstScores.GetValue()

class TestModelClass(ModelBase):
    def __init__(self, nScore: int = 4, *args, **kwargs) -> None:
        super().__init__()

        IntegerProperty(self, nScore, "_nScore")
        ModelProperty(self, TestModel(*args, **kwargs), "_mTestModel")

    @property
    def Score(self) -> int:
        return self._nScore.GetValue()

    @Score.setter
    def Score(self, nNewScore: int) -> None:
        self._nScore.SetValue(nNewScore)

    @property
    def TestModel(self) -> TestModel:
        return self._mTestModel.GetValue()


class TestModelListClass(ModelBase):
    def __init__(self, fScore: float = 8.5) -> None:
        super().__init__()

        FloatProperty(self, fScore, "_fScore")
        ModelListProperty(self, [], "_mModels")
        
    @property
    def Score(self) -> float:
        return self._fScore.GetValue()

    @Score.setter
    def Score(self, fNewScore: float) -> None:
        self._fScore.SetValue(fNewScore)

    @property
    def Models(self) -> List[TestModelClass]:
        return self._mModels.GetValue()

ModelBase.mSubModels[TestModel.__name__] = TestModel
ModelBase.mSubModels[TestModelClass.__name__] = TestModelClass
ModelBase.mSubModels[TestModelListClass.__name__] = TestModelListClass


from typing import *
import json
from ntt_signal import Signal
from .Exception import *
from .properties import *
from .IModel import IModel


class ModelBase(Signal, IModel):
    mSubModels: dict = {}

    def __init__(self) -> None:
        super().__init__()
        self._strDataAttributeNames: List[str] = []
        self._strModelDataAttributeNames: List[str] = []

    def _AppendModelData(self, strAttributeName: str, bIsModel: bool = False) -> None:
        if bIsModel:
            self._strModelDataAttributeNames.append(strAttributeName)
        else:
            self._strDataAttributeNames.append(strAttributeName)

    def ToDict(self) -> dict:
        dictData: dict = {}

        for strAttributeName in self._strDataAttributeNames:
            attribute: Property = getattr(self, strAttributeName)
            if isinstance(attribute, FloatProperty) \
                    or isinstance(attribute, IntegerProperty) \
                    or isinstance(attribute, ListProperty) \
                    or isinstance(attribute, StrProperty):
                dictData[strAttributeName] = attribute.GetValue()
            elif isinstance(attribute, ModelProperty):
                dictData[strAttributeName] = attribute.GetValue().ToDict()

        for strAttributeName in self._strModelDataAttributeNames:
            dictData[strAttributeName] = []
            mModels: List[IModel] = getattr(self, strAttributeName).GetValue()
            for mModel in mModels:
                dictData[strAttributeName].append(mModel.ToDict())


        dictData["__class__"] = self.__class__.__name__

        return dictData

    def FromDict(self, dictData: dict, bEmit: bool = True) -> None:
        if "__class__" not in dictData:
            raise InputDictError("Input dict has no __class__ key")
        elif dictData["__class__"] != self.__class__.__name__:
            raise TargetClassError(f"Expect {self.__class__.__name__} but received {dictData['__class__']}")
        else:
            for strAttributeName in self._strDataAttributeNames:
                if strAttributeName in dictData:
                    if not isinstance(dictData[strAttributeName], dict):
                        getattr(self,strAttributeName).SetValue(dictData[strAttributeName], False)
                    else:
                        getattr(self, strAttributeName).GetValue().FromDict(dictData[strAttributeName], False)

            for strAttributeName in self._strModelDataAttributeNames:
                if strAttributeName in dictData:
                    attribute: ModelListProperty = getattr(self, strAttributeName)
                    mModels = ObservableList([])
                    for dictObjectData in dictData[strAttributeName]:
                        obj: IModel = self.mSubModels[dictObjectData["__class__"]]()
                        obj.FromDict(dictObjectData, bEmit=False)

                        mModels.append(obj)

                    attribute.SetValue(mModels, False)

            if bEmit:
                self.Emit()

    def ToFile(self, strFileName: str) -> None:
        with open(strFileName, "w") as file:
            json.dump(self.ToDict(), file, indent=4)

    def FromFile(self, strFileName: str) -> None:
        with open(strFileName, "r") as file:
            self.FromDict(json.load(file))


from typing import Any
from ntt_signal import Signal

from ..IModel import IModel
from .Property import Property


class ModelProperty(Property):
    def __init__(self, model: IModel, mData: IModel, strAttributeName: str) -> None:
        super().__init__(model, strAttributeName)

        self._mData = mData
        self._mData.Attach(self)

    def GetValue(self) -> IModel:
        return self._mData
    
    def SetValue(self, mNewData: IModel, bNotify: bool = True) -> None:
        pass


from typing import Any
from ntt_signal import Signal
from .Property import Property


class FloatProperty(Property):
    def __init__(self, model: Signal, fValue: float, strAttributeName: str) -> None:
        super().__init__(model, strAttributeName, float)

        self._fValue = fValue

    def GetValue(self) -> Any:
        return self._fValue

    def SetValue(self, fNewValue: float, bNotify: bool = True) -> None:
        if isinstance(fNewValue, float) or isinstance(fNewValue, int):
            if fNewValue != self._fValue:
                self._fValue = float(fNewValue)
                if bNotify:
                    self.Emit(self._fValue)


from setuptools import setup, find_packages

def load_readme():
    with open('README.md', 'r', encoding='utf-8') as file:
        return file.read()

setup(
    name='ntt-json-model',
    version='1.0.0',
    packages=find_packages(),
    install_requires=[
    ],
    author='threezinedine',
    author_email='threezinedine@email.com',
    description='Small library for observation mechanism',
    long_description=load_readme(),
    long_description_content_type='text/markdown',
    url='https://github.com/threezinedine/ntt-json-model',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    keywords=['json', 'mvvm', 'signal', 'ntt'],
    project_urls={
        'Source': 'https://github.com/threezinedine/ntt-json-model',
        'Tracker': 'https://github.com/threezinedine/ntt-json-model/issues',
    },
)